from .random_nodes import Sampling
from .random_nodes import RandomNodes
from .random_edges import RandomEdges
from .degree_rank import DegreeRank
from .degree_group_rank import DegreeGroupRank
from .random_neighbor import RandomNeighbor
from .partial_crawls import PartialCrawls



from netin import PA
from netin import PAH
from netin import PATC
from netin import PATCH
from netin.utils import constants as const


class TestUnDiGraph(object):

    def test_patch_case_pa(self):
        n = 200
        k = 2
        f_m = 0.1
        seed = 5678
        g = PA(n=n, k=k, f_m=f_m, seed=seed)
        g.generate()
        c1 = not g.is_directed()
        c2 = g.number_of_nodes() == n
        c3 = g.calculate_minimum_degree() == k
        c4 = g.calculate_fraction_of_minority() == f_m
        c5 = g.get_model_name() == const.PA_MODEL_NAME
        assert c1 and c2 and c3 and c4 and c5, "Incorrect undirected parameters."

    def test_patch_case_pah(self):
        n = 200
        k = 2
        f_m = 0.1
        h_MM = 0.9
        h_mm = 0.1
        seed = 5678
        g = PAH(n=n, k=k, f_m=f_m, h_MM=h_MM, h_mm=h_mm, seed=seed)
        g.generate()
        c1 = not g.is_directed()
        c2 = g.number_of_nodes() == n
        c3 = g.calculate_minimum_degree() == k
        c4 = g.calculate_fraction_of_minority() == f_m
        c5 = g.get_model_name() == const.PAH_MODEL_NAME
        assert c1 and c2 and c3 and c4 and c5, "Incorrect undirected parameters."

    def test_case_patc(self):
        n = 200
        k = 2
        f_m = 0.1
        tc = 0.5
        seed = 5678
        g = PATC(n=n, k=k, f_m=f_m, tc=tc, seed=seed)
        g.generate()
        c1 = not g.is_directed()
        c2 = g.number_of_nodes() == n
        c3 = g.calculate_minimum_degree() == k
        c4 = g.calculate_fraction_of_minority() == f_m
        c5 = g.get_model_name() == const.PATC_MODEL_NAME
        assert c1 and c2 and c3 and c4 and c5, "Incorrect undirected parameters."

    def test_case_patch(self):
        n = 200
        k = 2
        f_m = 0.1
        h_MM = 0.1
        h_mm = 0.1
        tc = 1.0
        seed = 5678
        g = PATCH(n=n, k=k, f_m=f_m, h_MM=h_MM, h_mm=h_mm, tc=tc, seed=seed)
        g.generate()
        c1 = not g.is_directed()
        c2 = g.number_of_nodes() == n
        c3 = g.calculate_minimum_degree() == k
        c4 = g.calculate_fraction_of_minority() == f_m
        c5 = g.get_model_name() == const.PATCH_MODEL_NAME
        assert c1 and c2 and c3 and c4 and c5, "Incorrect undirected parameters."

    def test_case_all(self):
        n = 200
        k = 2
        f_m = 0.1
        h_MM = 0.5
        h_mm = 0.5
        tc = 0.0
        seed = 1234
        g_pa = PA(n=n, k=k, f_m=f_m, seed=seed)
        g_pah = PAH(n=n, k=k, f_m=f_m, h_MM=h_MM, h_mm=h_mm, seed=seed)
        g_patc = PATC(n=n, k=k, f_m=f_m, tc=tc, seed=seed)
        g_patch = PATCH(n=n, k=k, f_m=f_m, h_MM=h_MM, h_mm=h_mm, tc=tc, seed=seed)

        g_pa.generate()
        g_pah.generate()
        g_patc.generate()
        g_patch.generate()

        c1 = not g_pa.is_directed() and not g_pah.is_directed() and not g_patc.is_directed() and not g_patch.is_directed()
        c2 = g_pa.number_of_nodes() == g_pah.number_of_nodes() == g_patc.number_of_nodes() == g_patch.number_of_nodes() == n
        c3 = g_pa.calculate_minimum_degree() == g_pah.calculate_minimum_degree() == g_patc.calculate_minimum_degree() == g_patch.calculate_minimum_degree() == k
        c4 = g_pa.calculate_fraction_of_minority() == g_pah.calculate_fraction_of_minority() == g_patc.calculate_fraction_of_minority() == g_patch.calculate_fraction_of_minority() == f_m
        c5 = g_pa.get_expected_number_of_edges() == g_pah.get_expected_number_of_edges() == g_patc.get_expected_number_of_edges() == g_patch.get_expected_number_of_edges()
        c6 = g_pa.get_model_name() == const.PA_MODEL_NAME
        c7 = g_pah.get_model_name() == const.PAH_MODEL_NAME
        c8 = g_patc.get_model_name() == const.PATC_MODEL_NAME
        c9 = g_patch.get_model_name() == const.PATCH_MODEL_NAME
        assert c1 and c2 and c3 and c4 and c5 and c6 and c7 and c8 and c9, "Incorrect undirected parameters."



import logging
from typing import List
from typing import Set
from typing import Union

import matplotlib
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import rc

from netin.generators.graph import Graph
from netin.stats.distributions import fit_power_law
from netin.stats.distributions import get_disparity
from netin.stats.distributions import get_fraction_of_minority
from netin.stats.distributions import get_gini_coefficient
from netin.utils import constants as const
from netin.viz.constants import *
from netin.utils import validator as val


def reset_style():
    """
    Resets the style of the plots to the default style.

    """
    sns.reset_orig()


def set_paper_style(font_scale: float = 1.0):
    """
    Sets the style of the plots to the paper style.
    Font family serif, and allows to adjust the font scale.

    Parameters
    ----------
    font_scale: float
        A scale factor for the font size.

    """
    sns.set_context("paper", font_scale=font_scale)
    rc('font', family='serif')
    rc('lines', linewidth=1.0)
    rc('axes', linewidth=0.5)
    rc('xtick.major', width=0.5)
    rc('ytick.major', width=0.5)


def _get_edge_color(s: int, t: int, g: Graph) -> str:
    """
    Returns the color of the edge between s and t in the graph g.
    If the edge is homophilic (same attribute value for s and t), the color is the color of the attribute class.
    Otherwise, the color is mixed.

    Parameters
    ----------
    s: int
        source node
    t: int
        target node
    g:
        graph netin.Graph

    Returns
    -------
    color: str
        color of the edge
    """
    if g.get_class_value(s) == g.get_class_value(t):
        if g.get_class_value(s) == const.MINORITY_VALUE:
            return COLOR_MINORITY
        else:
            return COLOR_MAJORITY
    return COLOR_MIXED


def _get_class_label_color(class_label: str, ylabel: str = None) -> str:
    """
    Returns the color for the class label.

    Parameters
    ----------
    class_label: str
        class label e.g., M or m for majority and minority
    ylabel: str
        label of the y-axis.
        If it is a curve for the minority class, the color is the color of the minority class.
        Otherwise, the color for the class label.

    Returns
    -------
    color: str
        color for the class label
    """
    if ylabel in MINORITY_CURVE and class_label is None:
        return COLOR_MINORITY

    return COLOR_MINORITY if class_label == const.MINORITY_LABEL \
        else COLOR_MAJORITY if class_label == const.MAJORITY_LABEL \
        else COLOR_BLACK


def _save_plot(fig: matplotlib.figure.Figure, fn=None, **kwargs):
    """
    Saves the figure to a file, shows it and closes the figure.

    Parameters
    ----------
    fig: matplotlib.figure.Figure
        figure to save

    fn: str
        filename to save the figure

    kwargs: dict
        Additional arguments for the ``subplots_adjust`` function of the figure.

        left: float
            left margin

        right: float
            right margin

        bottom: float
            bottom margin

        top: float
            top margin

        wspace: float
            width space between subplots

        hspace: float
            height space between subplots

        Additional arguments for the ``savefig`` function of the figure.

        dpi: int
            dpi of the figure

        bbox_inches: str
            bounding box of the figure

    Returns
    -------

    """
    dpi = kwargs.pop('dpi', DPI)
    bbox_inches = kwargs.pop('bbox_inches', 'tight')
    wspace = kwargs.pop('wspace', None)
    hspace = kwargs.pop('hspace', None)
    left = kwargs.pop('left', None)
    right = kwargs.pop('right', None)
    bottom = kwargs.pop('bottom', None)
    top = kwargs.pop('top', None)

    fig.tight_layout()
    fig.subplots_adjust(left=left, right=right, bottom=bottom, top=top, wspace=wspace, hspace=hspace)
    if fn is not None and fig is not None:
        fig.savefig(fn, dpi=dpi, bbox_inches=bbox_inches, **kwargs)
        logging.info("%s saved" % fn)
    plt.show()
    plt.close()


def _get_grid_info(total_subplots: int, nc: int = None) -> (int, int):
    """
    Returns the number of columns and rows to be plotted
    based on the total number of subplots and the number of columns.

    Parameters
    ----------
    total_subplots: int
        total number of subplots
    nc: int
        number of columns

    Returns
    -------
    nc: int
        number of columns
    nr: int
        number of rows
    """
    nc = total_subplots if nc is None else nc  # min(MAX_PLOTS_PER_ROW, total_subplots) if nc is None else nc
    nr = int(np.ceil(total_subplots / nc))  # int(np.ceil(nc / MAX_PLOTS_PER_ROW))
    return nc, nr


def _add_class_legend(fig: matplotlib.figure.Figure, **kwargs):
    """
    Shows the legend of the figure using the "class_label" values of the nodes/graph.

    Parameters
    ----------
    fig: matplotlib.figure.Figure
        Figure to add the legend

    kwargs: dict
        Additional arguments for the ``legend`` function of the figure.

        bbox: tuple
            bounding box of the legend

        loc: str
            location of the legend
    """
    maj_patch = mpatches.Patch(color=COLOR_MAJORITY, label='majority')
    min_patch = mpatches.Patch(color=COLOR_MINORITY, label='minority')
    bbox = kwargs.pop('bbox', (1.04, 1))
    loc = kwargs.pop('loc', "upper left")
    fig.legend(handles=[maj_patch, min_patch], bbox_to_anchor=bbox, loc=loc)


def plot_graph(data: Union[Graph, Set[Graph], List[Graph]], share_pos: bool = False, fn: str = None, **kwargs):
    """
    Plots one or multiple (netin.Graph) graphs as matplotlib figures.

    Parameters
    ----------
    data: Union[netin.Graph, Set[netin.Graph]]
        graph or set of graphs to plot

    share_pos: bool
        if True, the positions of the nodes are shared between the graphs

    fn: str
        filename to save the figure

    kwargs: dict
        Additional arguments for the ``subplots`` function of the figure.

        cell_size: float
            size of the cell in inches

        nc: int
            number of columns

        node_size: int
            size of the nodes

        node_shape: str
            shape of the nodes e.g, 'o' for circle, 's' for square

        edge_width: float
            width of the edges

        edge_style: str
            style of the edges e.g., 'solid', 'dashed'

        edge_arrows: bool
            if True, the edges are plotted with arrows

        arrow_style: str
            style of the arrows e.g., '-\|>' for filled arrow

        arrow_size: int
            size of the arrows

        Additional arguments for the ``subplots_adjust`` and ``savefig`` functions of the figure.
    """
    iter_graph = [data] if isinstance(data, Graph) else data
    nc = kwargs.pop('nc', None)
    nc, nr = _get_grid_info(len(iter_graph), nc=nc)
    cell_size = kwargs.pop('cell_size', DEFAULT_CELL_SIZE)

    fig, axes = plt.subplots(nr, nc, figsize=(nc * cell_size, nr * cell_size), sharex=False, sharey=False)
    node_size = kwargs.get('node_size', 1)
    node_shape = kwargs.get('node_shape', 'o')
    edge_width = kwargs.get('edge_width', 0.02)
    edge_style = kwargs.get('edge_style', 'solid')
    edge_arrows = kwargs.get('edge_arrows', True)
    arrow_style = kwargs.get('arrow_style', '-|>')
    arrow_size = kwargs.get('arrow_size', 2)

    pos = None
    # same_n = len(set([g.number_of_nodes() for g in iter_graph])) == 1
    for cell in np.arange(nc * nr):
        row = cell // nc
        col = cell % nc
        ax = axes if nr == nc == 1 else axes[cell] if nr == 1 else axes[row, col]

        if cell < len(iter_graph):
            g = iter_graph[cell]

            pos = nx.spring_layout(g) if pos is None or not share_pos else pos # or not same_n

            # title
            ax.set_title(g.get_model_name())

            # nodes
            maj = g.graph['class_values'][g.graph['class_labels'].index("M")]
            nodes, node_colors = zip(
                *[(node, COLOR_MAJORITY if data[g.graph['class_attribute']] == maj else COLOR_MINORITY)
                  for node, data in g.nodes(data=True)])
            nx.draw_networkx_nodes(g, pos, nodelist=nodes, node_size=node_size, node_color=node_colors,
                                   node_shape=node_shape, ax=ax)

            # edges
            edges = g.edges()
            edges, edge_colors = zip(*[((s, t), _get_edge_color(s, t, g)) for s, t in edges])
            nx.draw_networkx_edges(g, pos, ax=ax, edgelist=edges, edge_color=edge_colors,
                                   width=edge_width, style=edge_style, arrows=edge_arrows, arrowstyle=arrow_style,
                                   arrowsize=arrow_size)

        # final touch
        ax.set_axis_off()

    # legend
    _add_class_legend(fig, **kwargs)
    _save_plot(fig, fn, **kwargs)


def plot_distribution(data: Union[pd.DataFrame, List[pd.DataFrame]],
                      col_name: Union[str, List],
                      get_x_y_from_df_fnc: callable, fn=None, **kwargs):
    """
    Plots a distribution of the values of a column of a dataframe.

    Parameters
    ----------
    data: Union[pandas.DataFrame, List[pandas.DataFrame]]
        Dataframe or list of dataframes to plot including. A column of this dataframe is used to plot the distribution.

    col_name: str
        name of the column to plot

    get_x_y_from_df_fnc: callable
        Function to get the x and y values from the dataframe

    fn: str
        Filename to save the figure

    kwargs: dict
        Additional arguments

        nc: int
            number of columns to show in the grid

        cell_size: float or tuple(float, float)
            size of the cell in inches. If a tuple is given, the first value is the width and the second value is the height.

        scatter: bool
            if True, the distribution is plotted as a scatter plot

        sharex: bool
            if True, the x-axis is shared between the subplots

        sharey: bool
            if True, the y-axis is shared between the subplots

        ylabel: str
            label of the y-axis if not set, the name of the function get_x_y_from_df_fnc is used

        xlabel: str
            label of the x-axis.

        xlim, ylim: tuple
            limits of the x and y-axis

        hue: str
            name of the column to use for grouping the data

        common_norm: bool
            If True, the y-axis is normalized to the same value (sum of all values). Otherwise, the y-axis is normalized per class (hue)

        log_scale: tuple(bool, bool)
            If True, the x and y-axis are plotted in log scale

        hline_fnc: callable
            function to plot a horizontal line

        vline_fnc: callable
            function to plot a vertical line

        suptitle: str
            title of the whole plot

        cuts: list
            list of cuts to plot (vertical lines)

        gini_fnc: callable
            function to plot the gini coefficient in ranking (inequality)

        me_fnc: callable
            function to plot the mean error in ranking (inequity)

        beta: float
            beta value for smoothing the inequity areas (over, under, and fair representation of minority nodes)

        class_label_legend: bool
            if True, the legend of the class labels is plotted

        And additional arguments sent to ``ax.scatter`` or ``ax.plot`` from matplotlib
    """
    iter_data = [data] if type(data) == pd.DataFrame else data
    nc = kwargs.pop('nc', None)
    nc, nr = _get_grid_info(len(iter_data), nc=nc)
    cell_size = kwargs.pop('cell_size', DEFAULT_CELL_SIZE)
    iter_column = [col_name] * (nc * nr) if type(col_name) == str else col_name

    # whole plot
    scatter = kwargs.pop('scatter', False)
    sharex = kwargs.pop('sharex', False)
    sharey = kwargs.pop('sharey', False)
    xy_fnc_name = get_x_y_from_df_fnc.__name__
    ylabel = xy_fnc_name.replace("get_", '').upper()
    ylabel = kwargs.pop('ylabel', ylabel)
    xlabel = kwargs.pop('xlabel', None)
    xlim = kwargs.pop('xlim', None)
    ylim = kwargs.pop('ylim', None)
    common_norm = kwargs.pop('common_norm', False)
    log_scale = kwargs.pop('log_scale', (False, False))
    hline_fnc = kwargs.pop('hline_fnc', None)
    vline_fnc = kwargs.pop('vline_fnc', None)
    wspace = kwargs.pop('wspace', None)
    hspace = kwargs.pop('hspace', None)
    suptitle = kwargs.pop('suptitle', None)

    # subplots
    cuts = kwargs.pop('cuts', None)
    gini_fnc = kwargs.pop('gini_fnc', None)
    me_fnc = kwargs.pop('me_fnc', None)
    beta = kwargs.pop('beta', None)

    # outter legend
    hue = kwargs.pop('hue', None)
    class_label_legend = kwargs.pop('class_label_legend', True)

    w, h = cell_size if type(cell_size) == tuple else (cell_size, cell_size)
    fig, axes = plt.subplots(nr, nc, figsize=(nc * w, nr * h), sharex=sharex, sharey=sharey)

    for cell, df in enumerate(iter_data):
        row = cell // nc
        col = cell % nc
        _col_name = iter_column[cell]

        ax = axes if nr == nc == 1 else axes[cell] if nr == 1 else axes[row, col]

        class_label: str
        iter_groups = df.groupby(hue) if hue is not None else [(None, df)]
        f_m = df.query("class_label == @const.MINORITY_LABEL").shape[0] / df.shape[0]
        for class_label, group in iter_groups:
            total = df[_col_name].sum() if common_norm else group[_col_name].sum()
            xs, ys = get_x_y_from_df_fnc(group, _col_name, total)
            plot = ax.scatter if scatter else ax.plot
            plot(xs, ys, label=class_label, color=_get_class_label_color(class_label, xy_fnc_name), **kwargs)

            if hline_fnc:
                hline_fnc(ax.axhline, group)
            if vline_fnc:
                vline_fnc(ax.axvline, group)
            if me_fnc:
                me_fnc(ax, f_m, ys, beta)
            if gini_fnc:
                gini_fnc(ax, ys, cuts)

        if log_scale[0]:
            ax.set_xscale('log')
        if log_scale[1]:
            ax.set_yscale('log')
        if xlim:
            ax.set_xlim(xlim)
        if ylim:
            ax.set_ylim(ylim)
        if ylabel and ((sharey and col == 0) or (not sharey)):
            ax.set_ylabel(ylabel)
        if xlabel is None:
            ax.set_xlabel(_col_name)
        elif (sharex and row == nr - 1) or (not sharex):
            ax.set_xlabel(xlabel)

        ax.set_title(df.name)

    # suptitle
    if suptitle is not None:
        fig.suptitle(suptitle)

    # legend
    if hue is not None and class_label_legend:
        _add_class_legend(fig, **kwargs)

    # save figure
    _save_plot(fig, fn, wspace=wspace, hspace=hspace, **kwargs)


def _show_cuts(axline, data):
    for c in const.INEQUALITY_CUTS:
        axline(c, ls='--', color='grey', alpha=0.5)


def _show_beta(axline, data):
    axline(const.INEQUITY_BETA, ls='--', color='grey', alpha=0.5)
    axline(-const.INEQUITY_BETA, ls='--', color='grey', alpha=0.5)


def plot_disparity(data: Union[pd.DataFrame, List[pd.DataFrame]], col_name: Union[str, List], fn: str = None, **kwargs):
    """Plots the disparity of the ranking of the minority group.

    Parameters
    ----------
    data: pd.DataFrame or List[pd.DataFrame]
        Metadata of nodes (each column is a property, e.g., degree)

    col_name: str
        Name of the column to plot (property of node)

    fn: str
        File name to save the plot

    kwargs: dict
        Additional arguments to pass to the ``plot_distribution`` function
    """
    gap = 0.04
    kwargs['class_label_legend'] = False
    kwargs['xlabel'] = INEQUITY_AXIS_LABEL
    kwargs['ylabel'] = INEQUALITY_AXIS_LABEL
    kwargs['ylim'] = (0.0 - gap, 1.0 + gap)
    kwargs['xlim'] = (-1.0 - gap, 1.0 + gap)
    kwargs['scatter'] = True
    kwargs['hline_fnc'] = _show_cuts
    kwargs['vline_fnc'] = _show_beta

    plot_distribution(data,
                      col_name=col_name,
                      get_x_y_from_df_fnc=get_disparity,
                      fn=fn, **kwargs)


def plot_gini_coefficient(data: Union[pd.DataFrame, List[pd.DataFrame]], col_name: Union[str, List],
                          fn: str = None, **kwargs):
    """
    Plots the Gini coefficient of the ranking of the minority group.

    Parameters
    ----------
    data: pd.DataFrame or List[pd.DataFrame]
        Metadata of nodes (each column is a property, e.g., degree)

    col_name: str or List[str]
        Name of the column to plot (property of node)

    fn: str
        File name to save the plot

    kwargs: dict
        Additional arguments to pass to the ``plot_distribution`` function
    """

    def show_gini(ax, ys, cuts):
        gini, x, y, va, ha, color = get_gini_label(ys, cuts)
        ax.text(s=gini,
                x=x, y=y,
                color=color,
                transform=ax.transAxes,
                ha=ha, va=va)

    def get_gini_label(ys, cuts=None) -> (str, float, str):
        from netin.stats import ranking
        # value
        gini_global = ys[-1]
        # label
        ineq = ranking.get_ranking_inequality_class(gini_global, cuts)
        # position
        right = False
        bottom = np.any(np.array(ys[:5]) > 0.8)
        y = 0.01 if bottom else 0.97
        x = 0.02
        va = 'bottom' if bottom else 'top'
        ha = 'right' if right else 'left'
        c = 'grey'
        return f"Gini={gini_global:.3f}\n{ineq}", x, y, va, ha, c

    cuts = kwargs.get('cuts', None)
    kwargs['cuts'] = cuts if cuts is not None else const.INEQUALITY_CUTS
    kwargs['class_label_legend'] = False
    kwargs['hline_fnc'] = _show_cuts
    kwargs['gini_fnc'] = show_gini
    kwargs['xlabel'] = RANKING_LABEL
    kwargs['ylabel'] = GINI_TOPK_AXIS_LABEL
    kwargs['ylim'] = (-0.01, 1.01)

    plot_distribution(data,
                      col_name=col_name,
                      get_x_y_from_df_fnc=get_gini_coefficient,
                      fn=fn, **kwargs)


def plot_fraction_of_minority(data: Union[pd.DataFrame, List[pd.DataFrame]], col_name: Union[str, List],
                              fn: str = None, **kwargs):
    """
    Plots the fraction of minority nodes in each top-k of the rank.

    Parameters
    ----------
    data: pd.DataFrame or List[pd.DataFrame]
        Metadata of nodes (each column is a property, e.g., degree)

    col_name: str or List
        Name of the column to plot (property of node)

    fn: str:
        File name to save the plot

    kwargs: dict
        Additional arguments to pass to the ``plot_distribution`` function
    """
    gap = 0.02

    def show_me(ax, f_m, ys, beta):
        me, x, y, va, ha, color = get_me_label(f_m, ys, beta)
        ax.text(s=me,
                x=x, y=y,
                color=color,
                transform=ax.transAxes,
                ha=ha, va=va)

    def get_me_label(f_m, ys, beta=None) -> (str, float, str):
        from netin.stats import ranking
        # value
        me = ranking.get_ranking_inequity(f_m, ys)
        # label
        ineq = ranking.get_ranking_inequity_class(me, beta)
        # position right
        right = False
        bottom = np.any(np.array(ys[:5]) > 0.8)
        y = 0 + (gap * 2) if bottom else 1 - (gap * 2)
        x = 0 + (gap * 2)

        if f_m <= 0.2 and bottom:
            # right - top
            right = True
            bottom = False
            y = 0 + (gap * 2) if bottom else 1 - (gap * 2)
            x = 1 - (gap * 2)

            if np.any(np.array(ys[5:]) > 0.8):
                # left - above fraction of minority
                right = False
                bottom = True
                y = f_m + (gap * 2)
                x = 0 + (gap * 2)

        va = 'bottom' if bottom else 'top'
        ha = 'right' if right else 'left'
        c = 'grey'
        return f"ME={me:.3f}\n{ineq}", x, y, va, ha, c

    def show_minority(axline, data):
        axline(data.query("class_label==@const.MINORITY_LABEL").shape[0] / data.shape[0], color="black", linestyle='--')

    beta = kwargs.get('beta', None)
    kwargs['beta'] = beta if beta is not None else const.INEQUITY_BETA
    kwargs['class_label_legend'] = False
    kwargs['hline_fnc'] = show_minority
    kwargs['me_fnc'] = show_me
    kwargs['xlabel'] = RANKING_LABEL
    kwargs['ylabel'] = FM_TOPK_AXIS_LABEL
    kwargs['ylim'] = (0. - gap, 1. + gap)

    # validation
    kwargs = val.ignore_params(['hue'], **kwargs)

    plot_distribution(data,
                      col_name=col_name,
                      get_x_y_from_df_fnc=get_fraction_of_minority,
                      fn=fn, **kwargs)


def plot_powerlaw_fit(data: Union[pd.DataFrame, List[pd.DataFrame]], col_name: Union[str, List], kind: str,
                      fn=None, **kwargs):
    """
    Plots the powerlaw fit of the data. It shows the empirical and fitted distributions

    Parameters
    ----------
    data: pd.DataFrame or List[pd.DataFrame]
        Metadata of nodes
    col_name: str or List[str]
        Column name of the data to plot (columns of the dataframe, for each (list) or for all (str)
    kind: str
        Type of distribution to plot. One of "ccdf", "cdf", "pdf"
    fn: str
        Name of the file to save the plot
    kwargs: dict
        Additional parameters to pass to the plot

        - nc: int
            Number of columns of the grid
        - cell_size: float or tuple(float, float)
            size of the cell in inches.
            If a tuple is given, the first value is the width and the second value is the height.
        - sharex: bool
            If True, the X axis will be shared among all subplots.
        - sharey: bool
            If True, the Y axis will be shared among all subplots.
        - log_scale: tuple(bool, bool)
            If True, the X and Y axis will be in log scale
        - xlabel: str
            Label of the X axis
        - ylabel: str
            Label of the Y axis
        - verbose: bool
            If True, it prints the parameters of the fit
        - wspace: float
            The amount of width reserved for blank space between subplots,
            expressed as a fraction of the average axis width
        - hspace: float
            The amount of height reserved for white space between subplots,
            expressed as a fraction of the average axis height
        - suptile: str
            Title of the whole plot
        - hue: str
            Column name of the data to use to color the plot (groups)
        - bbox: tuple(float, float, float, float)
            The bounding box that the whole plot will be fitted into.
            The four values represent (left, bottom, right, top) in figure coordinates.
        - loc: str
            Location of the legend. One of "best", "upper right", "upper left", "lower left", "lower right",
            "right", "center left", "center right", "lower center", "upper center", "center"

        Additional parameters to pass to the plot of each subplot (pdf, cdf, ccdf)

    """

    if kind not in TYPE_OF_DISTRIBUTION:
        raise ValueError(f"kind must be one of {TYPE_OF_DISTRIBUTION}")

    iter_data = [data] if type(data) == pd.DataFrame else data
    nc = kwargs.pop('nc', None)
    nc, nr = _get_grid_info(len(iter_data), nc=nc)
    cell_size = kwargs.pop('cell_size', DEFAULT_CELL_SIZE)
    iter_column = [col_name] * (nc * nr) if type(col_name) == str else col_name

    # whole plot
    sharex = kwargs.pop('sharex', False)
    sharey = kwargs.pop('sharey', False)
    log_scale = kwargs.pop('log_scale', (False, False))
    xlabel = kwargs.pop('xlabel', None)
    ylabel = kwargs.pop('ylabel',
                        "p(X≥x)" if kind == "ccdf" else "p(X<x)" if kind == 'cdf' else "p(X=x)" if kind == 'pdf' else None)
    verbose = kwargs.pop('verbose', False)
    wspace = kwargs.pop('wspace', None)
    hspace = kwargs.pop('hspace', None)
    suptitle = kwargs.pop('suptitle', None)

    # outer legend
    hue = kwargs.pop('hue', None)
    bbox = kwargs.pop('bbox', (1.0, 0.9))

    # inner legend
    fontsize = kwargs.pop('fontsize', None)
    loc = kwargs.pop('loc', None)

    # plot
    w, h = cell_size if type(cell_size) == tuple else (cell_size, cell_size)
    fig, axes = plt.subplots(nr, nc, figsize=(nc * w, nr * h), sharex=sharex, sharey=sharey)

    for cell, df in enumerate(iter_data):
        row = cell // nc
        col = cell % nc
        _col_name = iter_column[cell]

        ax = axes if nr == nc == 1 else axes[cell] if nr == 1 else axes[row, col]

        class_label: str
        iter_groups = df.groupby(hue) if hue is not None else [(None, df)]
        for class_label, group in iter_groups:
            group_nonzero = group.query(f"{_col_name}>0")

            if group_nonzero[_col_name].nunique() == 1:
                # not enough data to fit the powerlaw
                continue

            discrete = group_nonzero[_col_name].dtype == np.int64
            fit = fit_power_law(group_nonzero.loc[:, _col_name].values, discrete=discrete, verbose=verbose)

            color = _get_class_label_color(class_label)

            efnc = fit.plot_ccdf if kind == "ccdf" else fit.plot_cdf if kind == 'cdf' else fit.plot_pdf
            fnc = fit.power_law.plot_ccdf if kind == "ccdf" else fit.power_law.plot_cdf if kind == 'cdf' \
                else fit.power_law.plot_pdf

            try:
                ax = efnc(label=r"Empirical", ax=ax, color=color, **kwargs)
                ax = fnc(label=f'Powerlaw $\gamma={fit.alpha:.2f}$', linestyle='--', ax=ax, color=color, **kwargs)
            except Exception as ex:
                logging.warning("%s saved" % fn)

            # legend inside: empirical vs powerlaw
            handles, labels = ax.get_legend_handles_labels()
            loc = loc if loc is not None else 4 if kind == 'cdf' else 1
            leg = ax.legend(handles, labels, loc=loc, fontsize=fontsize)
            leg.draw_frame(False)

        if log_scale[0]:
            ax.set_xscale('log')
        if log_scale[1]:
            ax.set_yscale('log')
        if ylabel is not None and ((sharey and col == 0) or (not sharey)):
            ax.set_ylabel(ylabel)
        if xlabel is None:
            ax.set_xlabel(_col_name)
        elif (sharex and row == nr - 1) or (not sharex):
            ax.set_xlabel(xlabel)

        ax.set_title(df.name)

    # suptitle
    if suptitle is not None:
        fig.suptitle(suptitle)

    # legend (min, maj)
    if hue and bbox:
        _add_class_legend(fig, bbox=bbox, **kwargs)

    # save plot
    _save_plot(fig, fn, wspace=wspace, hspace=hspace, **kwargs)



############################################
# System dependencies
############################################
import operator
from typing import List
from typing import Tuple

import numpy as np

import netin
from . import constants as const
from .sampling import Sampling


############################################
# Class
############################################
class PartialCrawls(Sampling):
    """Sampling by partial crawls.

    Parameters
    ----------
    g: netin.Graph | netin.DiGraph
        global network

    pseeds: float
        fraction of seeds to sample

    max_tries: int
        maximum number of tries to sample a subgraph with enough classes and edges

    random_seed: object
        seed for random number generator

    kwargs: dict
        additional parameters for the sampling method

    References
    ----------
    .. [Yang2017] J. Yang, B. Ribeiro, & J. Neville "Should We Be Confident in Peer Effects Estimated From Social Network Crawls?" ICWSM (Vol. 11, No. 1, pp. 708-711), 2017.
    """

    ######################################################
    # Constructor
    ######################################################
    def __init__(self, g: netin.Graph, pseeds: float, max_tries: int = const.MAX_TRIES,
                 random_seed: object = None, **kwargs):
        super().__init__(g, pseeds, max_tries, random_seed, **kwargs)
        self.snsize = kwargs.get('snsize', const.SNSIZE)

    @property
    def method_name(self) -> str:
        return const.PARTIAL_CRAWLS

    def sampling(self):
        super().sampling()

    def _sample(self):
        """
        Creates a subgraph from G based on random edge sampling
        """
        num_classes = 0

        ### 1. pick random neighbors including source node both at random
        while num_classes < const.MIN_CLASSES:
            super_node = self._get_super_node()
            edges = self._get_edges_from_tours(super_node)
            nodes = set(np.array(list(edges)).flatten().tolist())
            num_classes = self._count_classes(nodes)

        return nodes, edges

    def _get_super_node(self) -> List[int]:
        """
        Randomly selects (sn * N) nodes from the graph `g`.

        Returns
        -------
        list
            list of nodes
        """
        nodes = list(self.g.nodes())
        np.random.shuffle(nodes)
        super_node_size = int(round(self.g.number_of_nodes() * self.snsize))
        sn = nodes[:super_node_size]
        return sn

    def _get_edges_from_tours(self, super_node: List[int]) -> List[Tuple[int, int]]:
        """
        Sorts nodes in super node proportional to their edges outside the super node.
        Performs random walks from/to super node until collecting (pseeds * N) nodes
        returns a list od edges

        Parameters
        ----------
        super_node: list
            list of nodes in the super node

        Returns
        -------
        list
            list of edges
        """
        # proportional to the number of edges out of the super node
        sorted_S = {ni: len([nj for nj in self.g.neighbors(ni) if nj not in super_node]) for ni in self.g.nodes()}
        sorted_S = sorted(sorted_S.items(), key=operator.itemgetter(1), reverse=True)
        sorted_S = [n[0] for n in sorted_S]

        # tours
        edges = set()
        sampled_nodes = set()
        for vi in sorted_S:
            nbrs_i = list(self.g.neighbors(vi))
            if len(nbrs_i) == 0:
                continue
            vj = np.random.choice(nbrs_i, 1)[0]  # random neighbor
            while vj not in super_node and len(sampled_nodes) < self.nseeds:
                edges.add((vi, vj))
                sampled_nodes |= {vi, vj}
                vi = vj
                nbrs_i = list(self.g.neighbors(vi))
                if len(nbrs_i) == 0:
                    break
                vj = np.random.choice(nbrs_i, 1)[0]  # random neighbor

        return edges



from typing import Union
from typing import Tuple
import warnings

import numpy as np
import networkx as nx
from collections import Counter

from netin.utils import constants as const
from netin.utils import validator as val


def get_min_degree(g: Union[nx.Graph, nx.DiGraph]) -> int:
    degrees = [d for n, d in g.degree]
    return min(degrees) if len(degrees) > 0 else -1


def _get_graph_metadata_value(g: Union[nx.Graph, nx.DiGraph], key: str, default: object = None) -> Union[object, iter]:
    value = default if key not in g.graph or g.graph[key] is None else g.graph[key]
    return value


def _get_class_labels(g: Union[nx.Graph, nx.DiGraph], class_attribute: str = None) -> Tuple[str, str, str]:
    if class_attribute:
        counter = Counter([obj[class_attribute] for n, obj in g.nodes(data=True)])
    else:
        val.validate_graph_metadata(g)
        class_attribute = _get_graph_metadata_value(g, 'class_attribute', const.CLASS_ATTRIBUTE)
        counter = Counter([obj[class_attribute] for n, obj in g.nodes(data=True)])

    if len(counter) > 2:
        warnings.warn(f'Graph contains more than two classes: {counter}')

    majority = counter.most_common()[0][0]
    minority = counter.most_common()[1][0]

    return majority, minority, class_attribute


def get_minority_fraction(g: Union[nx.Graph, nx.DiGraph], class_attribute: str = None) -> float:
    """
    Computes the fraction of the minority class in the graph.

    Parameters
    ----------
    g: Union[nx.Graph, nx.DiGraph]
        Graph to compute the fraction of the minority class

    """
    n = g.number_of_nodes()
    majority, minority, class_attribute = _get_class_labels(g, class_attribute)
    minority_count = sum([1 for n, obj in g.nodes(data=True) if obj[class_attribute] == minority])
    f_m = minority_count / n

    return f_m


def get_edge_type_counts(g: Union[nx.Graph, nx.DiGraph], fractions: bool = False, class_attribute: str = None) -> Counter:
    majority, minority, class_attribute = _get_class_labels(g, class_attribute)
    class_values = [majority, minority]
    class_labels = [const.MAJORITY_LABEL, const.MINORITY_LABEL]

    counts = Counter([f"{class_labels[class_values.index(g.nodes[e[0]][class_attribute])]}"
                      f"{class_labels[class_values.index(g.nodes[e[1]][class_attribute])]}"
                      for e in g.edges])

    if fractions:
        total = sum(counts.values())
        counts = Counter({k: v / total for k, v in counts.items()})

    return counts


def get_average_degree(g: Union[nx.Graph, nx.DiGraph]) -> float:
    k = sum([d for n, d in g.degree]) / g.number_of_nodes()
    return k


def get_average_degrees(g: Union[nx.Graph, nx.DiGraph], class_attribute: str = None) -> Tuple[float, float, float]:
    k = get_average_degree(g)

    majority, minority, class_attribute = _get_class_labels(g, class_attribute)
    kM = np.mean([d for n, d in g.degree if g.nodes[n][class_attribute] == majority])
    km = np.mean([d for n, d in g.degree if g.nodes[n][class_attribute] == minority])

    return k, kM, km


def get_similitude(g: Union[nx.Graph, nx.DiGraph], class_attribute: str = None) -> float:
    majority, minority, class_attribute = _get_class_labels(g, class_attribute)

    tmp = [int(g.nodes[e[0]][class_attribute] == g.nodes[e[1]][class_attribute]) for e in g.edges]
    total = len(tmp)
    sim = sum(tmp) / total

    return sim


def get_node_attributes(g: Union[nx.Graph, nx.DiGraph]) -> list:
    val.validate_graph_metadata(g)
    l = [a for n, a in nx.get_node_attributes(g, g.graph['class_attribute']).items()]
    return l



from typing import Union, Set, Tuple

import numpy as np

from netin.generators.h import Homophily
from netin.utils import constants as const
from .dpa import DPA


class DPAH(DPA, Homophily):
    """Creates a new DPAH instance. A directed graph with preferential attachment and homophily.

    Parameters
    ----------
    n: int
        number of nodes (minimum=2)

    d: float
        edge density (minimum=0, maximum=1)

    f_m: float
        fraction of minorities (minimum=1/n, maximum=(n-1)/n)

    plo_M: float
        activity (out-degree power law exponent) majority group (minimum=1)

    plo_m: float
        activity (out-degree power law exponent) minority group (minimum=1)

    h_MM: float
        homophily within majority group (minimum=0, maximum=1)

    h_mm: float
        homophily within minority group (minimum=0, maximum=1)

    seed: object
        seed for random number generator

    Notes
    -----
    The initialization is a directed with n nodes where f_m are the minority.
    Source nodes are selected based on their activity given by plo_M (if majority) or plo_m (if minority).
    Target nodes are selected via preferential attachment (in-degree) an homophily (h**).
    This model is based on [Espin-Noboa2022]_ which is the directed version of the "BA Homophily" model [Karimi2018]_.
    """

    ############################################################
    # Constructor
    ############################################################

    def __init__(self, n: int, d: float, f_m: float, plo_M: float, plo_m: float, h_MM: float, h_mm: float,
                 seed: object = None):
        DPA.__init__(self, n=n, d=d, f_m=f_m, plo_M=plo_M, plo_m=plo_m, seed=seed)
        Homophily.__init__(self, n=n, f_m=f_m, h_MM=h_MM, h_mm=h_mm, seed=seed)

    ############################################################
    # Init
    ############################################################

    def _infer_model_name(self):
        """
        Infers the name of the model.
        """
        return self.set_model_name(const.DPAH_MODEL_NAME)

    ############################################################
    # Generation
    ############################################################

    def _initialize(self, class_attribute: str = 'm', class_values: list = None, class_labels: list = None):
        """
        Initializes the model.

        Parameters
        ----------
        class_attribute: str
            name of the attribute that represents the class

        class_values: list
            values of the class attribute

        class_labels: list
            labels of the class attribute mapping the class_values.
        """
        DPA._initialize(self, class_attribute, class_values, class_labels)
        Homophily._initialize(self, class_attribute, class_values, class_labels)

    def get_target_probabilities(self, source: Union[None, int], target_set: Union[None, Set[int]],
                                 special_targets: Union[None, object, iter] = None) -> np.array:
        """
        Returns the probabilities of selecting a target node from a set of nodes based on
        preferential attachment and homophily, i.e., in-degree or target and homophily between source and target.

        Parameters
        ----------
        source: int
            source node

        target_set: Set[int]
            set of target nodes

        special_targets: object
            special targets

        Returns
        -------
        np.array
            probabilities of selecting a target node from a set of nodes
        """
        probs = np.array([self.get_homophily_between_source_and_target(source, target) *
                          (self.get_in_degree(target) + const.EPSILON) for target in target_set])
        probs /= probs.sum()
        return probs

    ############################################################
    # Calculations
    ############################################################

    def info_params(self):
        """
        Shows the (input) parameters of the model.
        """
        DPA.info_params(self)
        Homophily.info_params(self)

    def info_computed(self):
        """
        Shows the (computed) properties of the graph.
        """
        DPA.info_computed(self)
        Homophily.info_computed(self)

    def infer_homophily_values(self) -> Tuple[float, float]:
        """
        Infers the level of homophily within the majority and minority groups analytically.

        Returns
        -------
        Tuple[float, float]
            homophily within the majority and minority groups
        """
        h_MM, h_mm = None, None
        return h_MM, h_mm

    def _makecopy(self):
        """
        Makes a copy of the current object.
        """
        obj = self.__class__(n=self.n,
                              d=self.d,
                              f_m=self.f_m,
                              plo_M=self.plo_M,
                              plo_m=self.plo_m,
                              h_MM=self.h_MM,
                              h_mm=self.h_mm,
                              seed=self.seed)
        obj._initialize(class_attribute=self.class_attribute,
                        class_values=self.class_values,
                        class_labels=self.class_labels)
        return obj



############################################
# System dependencies
############################################
import operator
from typing import Tuple
from typing import Union

import netin
from . import constants as const
from .sampling import Sampling


############################################
# Class
############################################
class DegreeRank(Sampling):
    """Sampling by degree rank

    Parameters
    ----------
    g: netin.Graph | netin.DiGraph
        global network

    pseeds: float
        fraction of seeds to sample

    max_tries: int
        maximum number of tries to sample a subgraph with enough classes and edges

    random_seed: object
        seed for random number generator

    kwargs: dict
        additional parameters for the sampling method

        order: str
            order of nodes by degree. Options: "asc" | "desc"

    """

    ######################################################
    # Constructor
    ######################################################
    def __init__(self, g: netin.Graph, pseeds: float, random_seed: object = None, **kwargs):
        super().__init__(g=g, pseeds=pseeds, max_tries=1, random_seed=random_seed, **kwargs)
        self.order = self.kwargs.get("order", const.DESC)

    @property
    def method_name(self) -> str:
        name = f"{const.DEGREE_RANK} ({const.DESC if self.is_descending() else const.ASC})"
        return name

    def sampling(self):
        super().sampling()

    def is_ascending(self) -> bool:
        return not self.is_descending()

    def is_descending(self) -> bool:
        return self.order == const.DESC

    def _sample(self) -> Tuple[list, Union[list, None]]:
        """
        Creates a subgraph from G based on degree rank
        """
        edges = None

        ### 1. pick random nodes
        nodes = sorted([(n, d) for n, d in self.g.degree() if d > 0],
                       key=operator.itemgetter(1),
                       reverse=self.is_descending())
        nodes = nodes[:self.nseeds]
        nodes, degree = zip(*nodes)
        num_classes = self._count_classes(nodes)

        if num_classes < const.MIN_CLASSES:
            raise ValueError(f"{num_classes} class(es). Not enough classes in the sample of {self.nseeds} nodes."
                             "Try increasing the number of seeds or sampling by DegreeGroupRank.")

        return nodes, edges



import os
import glob
import pickle
import networkx as nx
from typing import Union, Tuple


def read_graph(fn: str) -> Union[nx.Graph, nx.DiGraph]:
    """
    Loads a graph from a file.

    Parameters
    ----------
    fn: str
        Path to file

    Returns
    -------
    Union[nx.Graph, nx.DiGraph]
        Graph
    """
    if fn.endswith('.gml'):
        return nx.read_gml(fn)
    elif fn.endswith('.pkl') or fn.endswith('.gpickle'):
        with open(fn, 'rb') as f:
            return pickle.load(f)
    else:
        raise ValueError(f'Unsupported file format: {fn}')



import time
from collections import Counter
from typing import Union, Set, List

import networkx as nx
import numpy as np
import pandas as pd
from pqdm.threads import pqdm
import warnings

import netin
from netin.utils import constants as const
from netin.utils import validator as val
from netin.stats import networks as net


class Graph(nx.Graph):
    """Bi-populated network (base graph model)

    Parameters
    ----------
    n: int
        number of nodes (minimum=2)

    f_m: float
        fraction of minorities (minimum=1/n, maximum=(n-1)/n)

    seed: object
        seed for random number generator

    Notes
    -----
    The initialization is a graph with n nodes and fraction of minority f_m.
    Source nodes are selected one-by-one (if undirected) and based on their activity (if directed).
    Target nodes are selected depending on the chosen edge mechanism.

    Available target mechanisms for undirected networks:
    - preferential attachment (pa), see :class:`netin.PA` [1]_.
    - preferential attachment with homophily (h_**), see :class:`netin.PAH` [2]_.
    - preferential attachment with triadic closure (tc), see :class:`netin.PATC` [3]_.
    - preferential attachment with homophily and triadic closure (tc), see :class:`netin.PATCH`.

    Available target mechanisms for directed networks:
    - preferential attachment, see :class:`netin.DPA` [4]_
    - homophily, see :class:`netin.DH` [4]_
    - preferential attachment with homophily (h_**), see :class:`netin.DPAH` [4]_

    References
    ----------
    .. [BarabasiAlbert1999] A. L. Barabasi and R. Albert "Emergence of scaling in random networks", Science 286, pp 509-512, 1999.
    .. [Karimi2018] F. Karimi, M. Génois, C. Wagner, P. Singer, & M. Strohmaier, M "Homophily influences ranking of minorities in social networks", Scientific reports 8(1), 11077, 2018.
    .. [HolmeKim2002] P. Holme and B. J. Kim “Growing scale-free networks with tunable clustering” Phys. Rev. E 2002.
    .. [Espin-Noboa2022] L. Espín-Noboa, C. Wagner, M. Strohmaier, & F. Karimi "Inequality and inequity in network-based ranking and recommendation algorithms" Scientific reports 12(1), 1-14, 2022.
    """

    ############################################################
    # Constructor
    ############################################################

    def __init__(self, n: int, f_m: float, seed: object = None):
        nx.Graph.__init__(self)
        self.n = n
        self.f_m = f_m
        self.seed = seed
        self.n_m = 0
        self.n_M = 0
        self.model_name = None
        self.class_attribute = None
        self.class_values = None
        self.class_labels = None
        self.node_list = None
        self.node_labels = None
        self._gen_start_time = None
        self._gen_duration = None

    ############################################################
    # Init
    ############################################################

    def _infer_model_name(self):
        """
        Infers the name of the model.
        """
        pass

    def _validate_parameters(self):
        """
        Validates the parameters of the graph (n, f_m, seed).
        """
        val.validate_int(self.n, minimum=2)
        val.validate_float(self.f_m, minimum=1 / self.n, maximum=(self.n - 1) / self.n)
        self.seed = self.seed if self.seed is not None else np.random.randint(0, 2 ** 32)

    def _set_class_info(self, class_attribute: str = 'm',
                        class_values: list = None,
                        class_labels: list = None):
        """
        Sets the class_attribute, class_values and class_labels.

        Parameters
        ----------
        class_attribute: str
            name of the class attribute

        class_values: list
            list of class values

        class_labels: list
            list of class labels
        """
        if class_labels is None:
            class_labels = const.CLASS_LABELS
        if class_values is None:
            class_values = const.CLASS_VALUES
        self.set_class_attribute(const.CLASS_ATTRIBUTE)
        self.set_class_values(class_values)
        self.set_class_labels(class_labels)

    def get_metadata_as_dict(self) -> dict:
        """
        Returns metadata for a graph as a dictionary.
        It includes the model name, class attribute, class values, class labels, number of nodes,
        fraction of minority, and seed.

        Returns
        -------
        dict
            metadata for a graph
        """
        obj = {'model': self.get_model_name(),
               'class_attribute': self.get_class_attribute(),
               'class_values': self.get_class_values(),
               'class_labels': self.get_class_labels(),
               'n': self.n,
               'f_m': self.f_m,
               'seed': self.seed}
        return obj

    ############################################################
    # Getters & Setters
    ############################################################

    def set_expected_number_of_nodes(self, n: int):
        """
        Sets the expected number of nodes.

        Parameters
        ----------
        n: int
            number of nodes
        """
        self.n = n

    def get_expected_number_of_nodes(self) -> int:
        """
        Returns the expected number of nodes (n, the input parameter).

        Returns
        -------
        int
            number of nodes
        """
        return self.n

    def get_expected_number_of_edges(self):
        pass

    def set_expected_fraction_of_minorities(self, f_m: float):
        """
        Sets the expected fraction of minorities.

        Parameters
        ----------
        f_m: float
            fraction of minorities (minimum=1/n, maximum=(n-1)/n)
        """
        self.f_m = f_m

    def get_expected_fraction_of_minorities(self) -> float:
        """
        Returns the expected fraction of minorities (f_m, the input parameter).

        Returns
        -------
        float
            fraction of minorities
        """
        return self.f_m

    def set_seed(self, seed=None):
        """
        Sets the random seed.

        Parameters
        ----------
        seed
            random seed
        """
        self.seed = seed

    def get_seed(self) -> object:
        """
        Returns the random seed (seed, the input parameter).

        Returns
        -------
        object
            random seed
        """
        return self.seed

    def set_model_name(self, model_name: str):
        """
        Sets the name of the model used to generate the graph.

        Parameters
        ----------
        model_name: str
            name of the model
        """
        self.model_name = model_name

    def get_model_name(self) -> str:
        """
        Returns the name of the model used to generate the graph.

        Returns
        -------
        str
            name of the model
        """
        return self.model_name

    def set_class_attribute(self, class_attribute: str):
        """
        Sets the class attribute of the graph used to infer minority and majority nodes.

        Parameters
        ----------
        class_attribute: str
            name of the class attribute
        """
        self.class_attribute = class_attribute

    def get_class_attribute(self) -> str:
        """
        Returns the class attribute of the graph used to infer minority and majority nodes.

        Returns
        -------
        str
            name of the class attribute
        """
        return self.class_attribute

    def set_class_values(self, class_values: list):
        """
        Sets the class values under the class attribute.
        These values are used to identify minority and majority nodes.
        
        Parameters
        ----------
        class_values: list
            list of class values
        """
        self.class_values = class_values

    def get_class_values(self) -> list:
        """
        Returns the class values under the class attribute.

        Returns
        -------
        list
            list of class values
        """
        return self.class_values

    def set_class_labels(self, class_labels: list):
        """
        Sets the class labels for minority and majority that map the class values.

        Parameters
        ----------
        class_labels: list
            list of class labels
        """
        self.class_labels = class_labels

    def get_class_labels(self) -> list:
        """
        Returns the class labels for minority and majority that map the class values.

        Returns
        -------
        list
            list of class labels
        """
        return self.class_labels

    def get_class_value(self, node: int) -> int:
        """
        Returns the class value of a given node (id).

        Parameters
        ----------
        node: int
            node id

        Returns
        -------
        int
            class value

        """
        return self.nodes[node][self.class_attribute]

    def get_class_label(self, node: int) -> str:
        """
        Returns the class label of a given node (id).

        Parameters
        ----------
        node: int
            node id

        Returns
        -------
        str
            class label
        """
        idx = self.class_values.index(self.nodes[node][self.class_attribute])
        return self.class_labels[idx]

    def get_majority_value(self) -> int:
        """
        Returns the value for the majority class.

        Returns
        -------
        int
            value for the majority class
        """
        return const.MAJORITY_VALUE

    def get_minority_value(self) -> int:
        """
        Returns the value for the minority class

        Returns
        -------
        int
            value for the minority class
        """
        return const.MINORITY_VALUE

    def get_majority_label(self) -> str:
        """
        Returns the label for the majority class.

        Returns
        -------
        str
            label for the majority class
        """
        return const.MAJORITY_LABEL

    def get_minority_label(self) -> str:
        """
        Returns the labels for the minority class.

        Returns
        -------
        str
            label for the minority class
        """
        return const.MINORITY_LABEL

    ############################################################
    # Generation
    ############################################################

    def _initialize(self, class_attribute: str = 'm', class_values: list = None, class_labels: list = None):
        """
        Initializes the random seed, the graph metadata, and node class information.
        """
        np.random.seed(self.seed)
        self._validate_parameters()
        self._init_graph(class_attribute, class_values, class_labels)
        self._init_nodes()

    def _init_graph(self, class_attribute: str = 'm', class_values: list = None, class_labels: list = None):
        """
        Initializes the graph.
        Sets the name of the model, class information, and the graph metadata.

        Parameters
        ----------
        class_attribute: str
            name of the class attribute

        class_values: list
            list of class values

        class_labels: list
            list of class labels
        """
        self._infer_model_name()
        self._set_class_info(class_attribute, class_values, class_labels)
        self.graph.update(self.get_metadata_as_dict())

    def _init_nodes(self):
        """
        Initializes the list of nodes with their respective labels.
        """
        self.node_list = np.arange(self.n)
        self.n_M = int(round(self.n * (1 - self.f_m)))
        self.n_m = self.n - self.n_M
        minorities = np.random.choice(self.node_list, self.n_m, replace=False)
        self.node_labels = {n: int(n in minorities) for n in self.node_list}

    def get_special_targets(self, source: int) -> object:
        pass

    def get_potential_nodes_to_connect(self, source: int, targets: Set[int]) -> Set[int]:
        """
        Returns the set of potential target nodes to connect to the source node.
        These target nodes are not the source node and are not already connected to the source node.

        Parameters
        ----------
        source: int
            source node id

        targets: Set[int]
            set of target node ids

        Returns
        -------
        Set[int]
            set of potential target node ids
        """
        return set([t for t in targets if t != source and t not in nx.neighbors(self, source)])

    def get_target_probabilities(self, source: Union[None, int], target_set: Union[None, Set[int]],
                                 special_targets: Union[None, object, iter] = None) -> tuple[np.array, set[int]]:
        """
        Returns the probability for each target node to be connected to the source node.

        Parameters
        ----------
        source:  int
            source node id (not used here)

        target_set: set
            set of target node ids

        special_targets:
            special targets to be considered (not used here)


        Notes
        -----
        It follows an Erdos-Renyi model [ErdosRenyi1959]_.

        Returns
        -------
        tuple[np.array, set[int]]
            tuple of probabilities and target set

        References
        ----------
        .. [ErdosRenyi1959] P. Erdős and A. Rényi, On Random Graphs, Publ. Math. 6, 290 (1959).
        """
        # Random (erdos renyi)
        probs = np.ones(len(target_set))
        probs /= probs.sum()
        return probs, target_set

    def get_target(self, source: Union[None, int], targets: Union[None, Set[int]],
                   special_targets: Union[None, object, iter]) -> int:
        pass

    def update_special_targets(self, idx_target: int, source: int, target: int, targets: Set[int],
                               special_targets: Union[None, object, iter]):
        pass

    def on_edge_added(self, source: int, target: int):
        pass

    def generate(self):
        """
        Basic instructions run before generating the graph.
        """
        self._gen_start_time = time.time()

        # init graph and nodes
        self._initialize()
        self.add_nodes_from(self.node_list)
        nx.set_node_attributes(self, self.node_labels, self.class_attribute)

        # iterate
        # for each source_node, get target (based on specific mechanisms), then add edge, update special targets

    def _terminate(self):
        """
        Instructions run after the generation of the graph.
        """
        self._gen_duration = time.time() - self._gen_start_time

    ############################################################
    # Calculations
    ############################################################

    def info_params(self):
        pass

    def info_computed(self):
        pass

    def info(self):
        """
        Shows the parameters and the computed propoerties of the graph.
        """
        print("=== Params ===")
        print('n: {}'.format(self.n))
        print('f_m: {}'.format(self.f_m))
        self.info_params()
        print('seed: {}'.format(self.seed))

        print("=== Model ===")
        print('Model: {}'.format(self.get_model_name()))
        print('Class attribute: {}'.format(self.get_class_attribute()))
        print('Class values: {}'.format(self.get_class_values()))
        print('Class labels: {}'.format(self.get_class_labels()))
        print('Generation time: {} (secs)'.format(self._gen_duration))

        print("=== Computed ===")
        print(f'- is directed: {self.is_directed()}')
        print(f'- number of nodes: {self.number_of_nodes()}')
        print(f'- number of edges: {self.number_of_edges()}')
        print(f'- minimum degree: {self.calculate_minimum_degree()}')
        print(f'- fraction of minority: {self.calculate_fraction_of_minority()}')
        print(f'- edge-type counts: {self.calculate_edge_type_counts()}')
        print(f"- density: {nx.density(self)}")
        try:
            print(f"- diameter: {nx.diameter(self)}")
        except Exception as ex:
            print(f"- diameter: <{ex}>")
        try:
            print(f"- average shortest path length: {nx.average_shortest_path_length(self)}")
        except Exception as ex:
            print(f"- average shortest path length: <{ex}>")
        print(f"- average degree: {net.get_average_degree(self)}")
        print(f"- degree assortativity: {nx.degree_assortativity_coefficient(self)}")
        print(f"- attribute assortativity ({self.class_attribute}): "
              f"{nx.attribute_assortativity_coefficient(self, self.class_attribute)}")
        print(f"- transitivity: {nx.transitivity(self)}")
        print(f"- average clustering: {nx.average_clustering(self)}")
        self.info_computed()

    def calculate_minimum_degree(self) -> int:
        """
        Returns the minimum degree of the graph.

        Returns
        -------
        int
            minimum degree
        """
        return net.get_min_degree(self)

    def calculate_fraction_of_minority(self) -> float:
        """
        Returns the fraction of minority nodes in the graph (based on class attribute).

        Returns
        -------
        float
            fraction of minority nodes
        """
        return net.get_minority_fraction(self)

    def calculate_edge_type_counts(self) -> Counter:
        """
        Returns the number of edges of each type, e.g., Mm (between majority and minority), mm, etc.

        Returns
        -------
        Counter
            counter of edges types
        """
        return net.get_edge_type_counts(self)

    ############################################################
    # Metadata
    ############################################################

    def compute_node_stats(self, metric: str, **kwargs) -> List[Union[int, float]]:
        """
        Returns the property of each node in the graph based on the metric.

        Parameters
        ----------
        metric: str
            metric to compute

        kwargs: dict
            additional parameters for the metric

        Returns
        -------
        List[Union[int, float]]
            list properties for each node
        """
        values = None

        # list of tuples (node, value)
        if metric == 'degree':
            values = self.degree(self.node_list, **kwargs) if not self.is_directed() else None
        if metric == 'in_degree':
            values = self.in_degree(self.node_list, **kwargs) if self.is_directed() else None
        if metric == 'out_degree':
            values = self.out_degree(self.node_list, **kwargs) if self.is_directed() else None
        if metric == 'eigenvector':
            try:
                values = nx.eigenvector_centrality_numpy(self, **kwargs)
            except Exception as ex:
                try:
                    values = nx.eigenvector_centrality_numpy(self, max_iter=200, tol=1.0e-5)
                except Exception as ex:
                    warnings.warn(f"The eigenvector centrality could not be computed: {ex}")
                    values = None

        # dict of node -> value
        if metric == 'clustering':
            values = nx.clustering(self, self.node_list, **kwargs)
        if metric == 'betweenness':
            values = nx.betweenness_centrality(self, **kwargs)
        if metric == 'closeness':
            if isinstance(self, nx.DiGraph):
                values = nx.closeness_centrality(nx.DiGraph(self), **kwargs)
            else:
                values = nx.closeness_centrality(self, **kwargs)
        if metric == 'pagerank':
            values = nx.pagerank(self, **kwargs)

        return [values[n] for n in self.node_list] if values is not None else np.nan

    def get_node_metadata_as_dataframe(self, include_graph_metadata: bool = False, n_jobs: int = 1) -> pd.DataFrame:
        """
        Returns the metadata of the nodes in the graph as a dataframe.
        Every row represents a node, and the columns are the metadata of the node.

        Parameters
        ----------
        include_graph_metadata: bool
            whether to include the graph metadata (e.g., class attribute, class values, etc.)

        n_jobs: int
            number of parallel jobs

        Returns
        -------
        pd.DataFrame
            dataframe with the metadata of the nodes
        """
        cols = ['node', 'class_label']
        obj = {'node': self.node_list,
               'class_label': [self.get_class_label(n) for n in self.node_list]}

        # include graph metadata
        if include_graph_metadata:
            n = self.number_of_nodes()
            newcols = [c for c in self.graph.keys() if c not in ['class_attribute', 'class_values', 'class_labels']]
            obj.update({c: self.graph[c] for c in newcols})
            cols.extend(newcols)

        # include metrics
        column_values = pqdm(const.VALID_METRICS, self.compute_node_stats, n_jobs=n_jobs)
        obj.update({col: values for col, values in zip(const.VALID_METRICS, column_values)})
        cols.extend(const.VALID_METRICS)

        # create dataframe
        df = pd.DataFrame(obj, columns=cols, index=self.node_list)
        df.name = self.get_model_name()

        # add ranking values
        for metric in const.VALID_METRICS:
            ncol = f'{metric}_rank'
            df.loc[:, ncol] = df.loc[:, metric].rank(ascending=False, pct=True, method='dense')

            # # compute ranking values and retry for ARPACK error.
            # done = False
            # tries = 10
            # while not done:
            #     try:
            #         df.loc[:, ncol] = df.loc[:, metric].rank(ascending=False, pct=True, method='dense')
            #         done = True
            #     except Exception as ex:
            #         tries -= 1
            #         if tries <= 0:
            #             raise UserWarning(f"An error occurred while computing the ranking values: {ex}")

        return df

    def _makecopy(self) -> Union[nx.Graph, nx.DiGraph]:
        pass

    def copy(self) -> Union[nx.Graph, nx.DiGraph]:
        """
        Makes a copy of the current object.
        Returns
        -------
        netin.Graph
            copy of the current object
        """
        g = self._makecopy()
        g._init_graph(class_attribute=self.class_attribute,
                      class_values=self.class_values,
                      class_labels=self.class_labels)
        g.graph.update(self.graph)
        g.add_nodes_from((n, d.copy()) for n, d in self._node.items())
        g.add_edges_from(
            (u, v, datadict.copy())
            for u, nbrs in self._adj.items()
            for v, datadict in nbrs.items()
        )
        g.node_list = self.node_list.copy()
        g.node_labels = self.node_labels.copy()
        g.n_m = self.n_m
        g.n_M = self.n_M
        return g



# from .sampling.random_nodes import RandomNodes
# from .sampling.random_edges import RandomEdges
# from .sampling.degree_rank import DegreeRank
# from .sampling.degree_group_rank import DegreeGroupRank
# from .sampling.random_neighbor import RandomNeighbor
# from .sampling.partial_crawls import PartialCrawls



from netin import DPA


def run():
    n = 200
    d = 0.1
    f_m = 0.1
    plo_M = 2.0
    plo_m = 2.0
    seed = 1234
    g = DPA(n=n, d=d, f_m=f_m, plo_M=plo_M, plo_m=plo_m, seed=seed)
    g.generate()
    g.info()


if __name__ == '__main__':
    run()



MAX_PLOTS_PER_ROW = 5
COLOR_MAJORITY = 'tab:blue'
COLOR_MINORITY = 'tab:orange'
COLOR_MIXED = 'tab:green'
COLOR_BLACK = 'black'
INACTIVE_COLOR = 'lightgrey'
RANKING_LABEL = 'rank k'
FM_TOPK_AXIS_LABEL = 'f$_m$ in top-k'
GINI_TOPK_AXIS_LABEL = 'Gini in top-k'
INEQUITY_AXIS_LABEL = 'ME'
INEQUALITY_AXIS_LABEL = 'Gini'
DPI = 300
DEFAULT_CELL_SIZE = 3
DEFAULT_FIGSIZE = (10, 5)
MINORITY_CURVE = ['get_fraction_of_minority']
TYPE_OF_DISTRIBUTION = ['pdf', 'cdf', 'ccdf']


from typing import Union, Set, List, Tuple

import numpy as np
import pandas as pd

from netin.utils import constants as const


### Inequity (fraction of minority in ranking) ###

def get_ranking_inequity(f_m: float, ys: np.array) -> float:
    """Computes ME: mean error distance between the fraction of minority in each top-k rank `f_m^k` and
    the fraction of minority of the entire graph `f_m`. ME is the ranking inequity of the rank.

    Parameters
    ----------
    f_m: float
        The fraction of minority in the entire graph.

    ys: np.array
        The fraction of minority in each top-k rank.

    Returns
    -------
    me: float
        The ranking inequity of the rank.
    """
    me = np.mean([efm - f_m for efm in ys if not np.isnan(efm)])
    return me


def get_ranking_inequity_class(me: float, beta: float = None) -> str:
    """
    Infers the inequity class (label) given the inequity measure (ME).

    Parameters
    ----------
    me: float
        The inequity measure (ME).

    beta: float
        The threshold to determine the inequity class.

    Returns
    -------
    label: str
        The inequity class label (i.e., fair, over-represented, under-represented).

    Notes
    -----
    See :func:`get_ranking_inequity` for more details on `me`.

    By default, `beta=0.05`, see [Espin-Noboa2022].
    """
    beta = const.INEQUITY_BETA if beta is None else beta
    label = const.INEQUITY_OVER if me > beta else const.INEQUITY_UNDER if me < -beta else const.INEQUITY_FAIR
    return label


def get_fraction_of_minority_in_ranking(df: pd.DataFrame, x: str) -> \
        Union[Tuple[np.ndarray, np.ndarray], Tuple[list, list]]:
    """
    Computes the fraction of minority in each top-k rank.

    Parameters
    ----------
    df: pd.DataFrame
        DataFrame that contains the data.

    x: str
        The column name of the data.

    Returns
    -------
    xs: np.ndarray
        The x values (ranking).

    ys: np.ndarray
        The y values (fraction of minority).
    """
    xs = const.RANK_RANGE
    ys = []
    for rank in xs:
        column = f"{x}_rank"
        tmp = df.query(f"{column} <= @rank").copy()
        total = tmp.shape[0]
        efm = np.nan if total == 0 else tmp.query("class_label == @const.MINORITY_LABEL").shape[0] / total
        ys.append(efm)
    return xs, ys


### Inequality (Gini coefficient) ###

def get_ranking_inequality(ys: np.array) -> float:
    """
    Returns the Gini coefficient of the entire distribution (at op-100%).

    Parameters
    ----------
    ys: np.array
        The y values (Gini coefficients in each top-k rank).

    Returns
    -------
    float
        The Gini coefficient of the entire distribution (at op-100%).
    """
    gini_global = ys[0]  # top-100%
    return gini_global


def get_ranking_inequality_class(gini_global: float, cuts: Set[float] = const.INEQUALITY_CUTS) -> str:
    """
    Infers the inequality class label given the Gini coefficient of the entire distribution.

    Parameters
    ----------
    gini_global: float
        The Gini coefficient of the entire distribution.

    cuts: Set[float]
        The cuts to determine the inequality class.

    Returns
    -------
    label: str
        The inequality class label (i.e., equality, moderate, skewed)

    Notes
    -----
    By default, `cuts={0.3, 0.6}`, see [Espin-Noboa2022].
    """
    cuts = const.INEQUALITY_CUTS if cuts is None else cuts
    if len(cuts) != 2 or len(set(cuts)) == 1:
        raise Exception("There must be two cuts for the inequality class")

    label = const.INEQUALITY_HIGH if gini_global >= max(cuts) \
        else const.INEQUALITY_LOW if gini_global <= min(cuts) \
        else const.INEQUALITY_MODERATE
    return label


def get_gini_in_ranking(df: pd.DataFrame, x: str) -> Union[Tuple[np.ndarray, np.ndarray], Tuple[list, list]]:
    """
    Computes the Gini coefficient of a distribution `df[x]` in each top-k rank.

    Parameters
    ----------
    df: pd.DataFrame
        Dataframe that contains the data.

    x: str
        The column name of the data.

    Returns
    -------
    xs: np.ndarray
        The x values (ranking).
    ys: np.ndarray
        The y values (Gini coefficients).
    """
    xs = const.RANK_RANGE
    ys = []
    for rank in xs:
        column = f"{x}_rank"
        tmp = df.query(f"{column} <= @rank").copy()
        g = gini(tmp.loc[:, x].values)
        ys.append(g)
    return xs, ys


def gini(data: np.array) -> float:
    """
    Calculates the Gini coefficient of a distribution.

    Parameters
    ----------
    data: np.array
        The data.

    Returns
    -------
    float
        The Gini coefficient of the distribution.

    References
    ----------
    `Gini coefficient <http://www.statsdirect.com/help/default.htm#nonparametric_methods/gini.htm>`_
    `Implementation <https://github.com/oliviaguest/gini/blob/master/gini.py>`_
    """
    # https://github.com/oliviaguest/gini/blob/master/gini.py
    # based on bottom eq:
    # http://www.statsdirect.com/help/generatedimages/equations/equation154.svg
    # from:
    # http://www.statsdirect.com/help/default.htm#nonparametric_methods/gini.htm
    # All values are treated equally, arrays must be 1d:
    X = data.flatten().astype(np.float64)
    if np.amin(X) < 0:
        # Values cannot be negative:
        X -= np.amin(X)
    # Values cannot be 0:
    X += 0.0000001
    # Values must be sorted:
    X = np.sort(X)
    # Index per array element:
    index = np.arange(1, X.shape[0] + 1)
    # Number of array elements:
    n = X.shape[0]
    # Gini coefficient:
    return (np.sum((2 * index - n - 1) * X)) / (n * np.sum(X))



from .distributions import get_pdf
from .distributions import get_cdf
from .distributions import get_ccdf
from .distributions import fit_power_law
from .distributions import get_fraction_of_minority
from .distributions import get_gini_coefficient
from .distributions import get_disparity
from .networks import get_min_degree
from .networks import get_minority_fraction
from .networks import get_edge_type_counts
from .networks import get_average_degree
from .networks import get_average_degrees



RANDOM_NODES = "Random node sampling"
RANDOM_EDGES = "Random edge sampling"
DEGREE_RANK = "Degree Rank"
DEGREE_GROUP_RANK = "Degree Group Rank"
RANDOM_NEIGHBORS = "Random neighbor sampling"
PARTIAL_CRAWLS = "Partial crawl sampling"

DESC = "desc"
ASC = "asc"

SNSIZE = 0.01

MAX_TRIES = 20
MIN_EDGES = 2
MIN_CLASSES = 2



from .constants import COLOR_MAJORITY
from .constants import COLOR_MINORITY
from .constants import COLOR_MIXED
from .handlers import plot_graph
from .handlers import plot_distribution
from .handlers import plot_powerlaw_fit
from .handlers import plot_fraction_of_minority
from .handlers import plot_gini_coefficient
from .handlers import plot_disparity
from .handlers import reset_style
from .handlers import set_paper_style



from typing import Set
from typing import Union

import numpy as np

from netin.utils import constants as const
from netin.generators.undirected import UnDiGraph


class PA(UnDiGraph):
    """Creates a new PA instance. An undirected graph with preferential attachment.

    Parameters
    ----------
    n: int
        number of nodes (minimum=2)

    k: int
        minimum degree of nodes (minimum=1)

    f_m: float
        fraction of minorities (minimum=1/n, maximum=(n-1)/n)

    seed: object
        seed for random number generator

    Notes
    -----
    The initialization is an undirected graph with n nodes and no edges.
    Then, everytime a node is selected as source, it gets connected to k target nodes.
    Target nodes are selected via preferential attachment (in-degree), see [BarabasiAlbert1999]_.
    """
    ############################################################
    # Constructor
    ############################################################

    def __init__(self, n: int, k: int, f_m: float, seed: object = None):
        super().__init__(n, k, f_m, seed)

    ############################################################
    # Init
    ############################################################

    def _infer_model_name(self):
        """
        Infers the name of the model.
        """
        return self.set_model_name(const.PA_MODEL_NAME)

    ############################################################
    # Generation
    ############################################################

    def get_target_probabilities(self, source: Union[None, int], target_set: Union[None, Set[int]],
                                 special_targets: Union[None, object, iter] = None) -> tuple[np.array, set[int]]:
        """
        Returns the probabilities of the target nodes to be selected given a source node.
        This probability is proportional to the degree of the target node.

        Parameters
        ----------
        source: int
            source node (id)

        target_set: set
            set of target nodes (ids)

        special_targets: object
            special targets

        Returns
        -------
        probs: np.array
            probabilities of the target nodes to be selected

        target_set: set
            set of target nodes (ids)
        """
        probs = np.array([(self.degree(target) + const.EPSILON) for target in target_set])
        probs /= probs.sum()
        return probs, target_set

    def _makecopy(self):
        """
        Makes a copy of the current object.
        """
        return self.__class__(n=self.n,
                              k=self.k,
                              f_m=self.f_m,
                              seed=self.seed)
